/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import com.bitctrl.text.CheckText;

/**
 * Ein Fuzzy-Set bestehend aus einem Term (der Name) und der Definition der
 * Zugeh&ouml;rigkeitsfunktion. Optional k&ouml;nnen die &Uuml;berg&auml;nge
 * zwischen Fuzzy-Sets automatisch in der linguistischen Variable berechnet
 * werden.
 * <p>
 * Werden die erste und letzte Trapezecke nicht explizit gesetzt, stehen sie auf
 * {@link java.lang.Integer#MIN_VALUE} bzw. {@link java.lang.Integer#MAX_VALUE},
 * was als Flag f&uuml;r das Berechnen des &Uuml;bergangs interpretiert wird.
 * </p>
 * <p>
 * Die Trapezecken m&uuml;ssen auf der x-Achse aufsteigend sortiert angegeben
 * werden. Einzige Ausnahme ist ein Fuzzy-Set, welches die Wertebereichsgrenzen
 * &uuml;berdeckt, hier gilt: t4&lt;t1, aber auch t3&lt;t4 und t1&lt;t2.
 * </p>
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class FuzzySet {

	/**
	 * X-Koordinate der ersten Trapezecke, die y-Koordinate ist immer 0.
	 */
	private int t1;

	/**
	 * X-Koordinate der zweiten Trapezecke, die y-Koordinate ist immer 1.
	 */
	private int t2;

	/**
	 * X-Koordinate der dritten Trapezecke, die y-Koordinate ist immer 1.
	 */
	private int t3;

	/**
	 * X-Koordinate der vierten Trapezecke, die y-Koordinate ist immer 0.
	 */
	private int t4;

	/**
	 * Name des Fuzzy-Sets bzw. Terms
	 */
	private final String name;

	/**
	 * Konstruiert ein Fuzzy-Set. Die fehlenden Ecken t1 und t4 werden
	 * sp&auml;ter berechnet. Ruft intern
	 * {@link FuzzySet#FuzzySet(String, int, int, int, int)} mit t1 =
	 * {@link java.lang.Integer#MIN_VALUE} und t4 =
	 * {@link java.lang.Integer#MAX_VALUE} auf.
	 *
	 * @param name
	 *            Name des Terms
	 * @param t2
	 *            Zweite Ecke des Trapez
	 * @param t3
	 *            Dritte Ecke des Trapez
	 */
	public FuzzySet(final String name, final int t2, final int t3) {
		this(name, Integer.MIN_VALUE, t2, t3, Integer.MAX_VALUE);
	}

	/**
	 * Konstruiert ein Fuzzy-Set.
	 *
	 * @param name
	 *            Name des Terms
	 * @param t1
	 *            Erste Ecke des Trapez
	 * @param t2
	 *            Zweite Ecke des Trapez
	 * @param t3
	 *            Dritte Ecke des Trapez
	 * @param t4
	 *            Vierte Ecke des Trapez
	 */
	public FuzzySet(final String name, final int t1, final int t2, final int t3,
			final int t4) {
		if (!CheckText.isPrintable(name) || CheckText.containsLineBreak(name)) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.TermnameUngueltig.toString());
		}

		if (isSortiert(t1, t2, t3, t4) || isSpezialfall(t1, t2, t3, t4)) {
			this.name = name.trim();
			this.t1 = t1;
			this.t2 = t2;
			this.t3 = t3;
			this.t4 = t4;
		} else {
			throw new IllegalArgumentException(
					FuzzyLibMsg.TrapezeckenUnsortiert.toString());
		}
	}

	/**
	 * Zwei Fuzzy-Sets sind "gleich", wenn sie den gleichen Termnamen haben.
	 * Dieses Verhalten ist f&uuml;r das Hashing notwendig.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(final Object obj) {
		if (obj == this) {
			return true;
		} else if (obj instanceof FuzzySet) {
			final FuzzySet fs = (FuzzySet) obj;
			return name.equals(fs.name);
		}

		return false;
	}

	@Override
	public int hashCode() {
		return name.hashCode();
	}

	/**
	 * Gibt den Name des Terms zur&uuml;ck.
	 *
	 * @return Termname
	 */
	public String getName() {
		return name;
	}

	/**
	 * Gibt die erste Ecke des Trapez zur&uuml;ck.
	 *
	 * @return Erste Ecke des Trapez
	 */
	public int getT1() {
		return t1;
	}

	/**
	 * Gibt die zweite Ecke des Trapez zur&uuml;ck.
	 *
	 * @return Zweite Ecke des Trapez
	 */
	public int getT2() {
		return t2;
	}

	/**
	 * Gibt die dritte Ecke des Trapez zur&uuml;ck.
	 *
	 * @return Dritte Ecke des Trapez
	 */
	public int getT3() {
		return t3;
	}

	/**
	 * Gibt die vierte Ecke des Trapez zur&uuml;ck.
	 *
	 * @return Vierte Ecke des Trapez
	 */
	public int getT4() {
		return t4;
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set eine absteigende Rampe ist. Die Rampe muss
	 * nicht an das Minimum des Wertebereichs angrenzen!
	 *
	 * @return {@code true}, wenn das Fuzzy-Set eine absteigende Rampe darstellt
	 */
	public boolean isAbsteigendeRampe() {
		if (isSpezialfall()) {
			return false;
		}

		return (t1 == t2) && (t2 < t3) && (t3 < t4);
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set eine aufsteigende Rampe ist. Die Rampe muss
	 * nicht an das Maximum des Wertebereichs angrenzen!
	 *
	 * @return {@code true}, wenn das Fuzzy-Set eine aufsteigende Rampe
	 *         darstellt
	 */
	public boolean isAufsteigendeRampe() {
		if (isSpezialfall()) {
			return false;
		}

		return (t1 < t2) && (t2 < t3) && (t3 == t4);
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set ein Dreieck ist.
	 *
	 * @return {@code true}, wenn das Fuzzy-Set ein Dreieck darstellt
	 */
	public boolean isDreieck() {
		if (isSpezialfall()) {
			return false;
		}

		return (t1 < t2) && (t2 == t3) && (t3 < t4);
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set ein Spezialfall ist. Es gibt genau einen
	 * Spezialfall und zwar, wenn das Fuzzy-Sets beide Wertebereichsgrenzen
	 * schneidet. Dies ist der einzige erlaubte Fall, in dem die Trapezecken
	 * nicht aufsteigend sortiert sind. Die aufsteigende Kante t1-t2 liegt in
	 * diesem Fall hinter der absteigenden Kante t3-t4.
	 *
	 * @return {@code true}, wenn der Spezialfall vorliegt
	 */
	public boolean isSpezialfall() {
		return ((t1 < t2) && (t2 > t3) && (t3 < t4));
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set eine Spitze ist.
	 *
	 * @return {@code true}, wenn das Fuzzy-Set eine Spitze darstellt
	 */
	public boolean isSpitze() {
		if (isSpezialfall()) {
			return false;
		}

		return (t1 == t2) && (t2 == t3) && (t3 == t4);
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set ein Trapez ist.
	 *
	 * @return {@code true}, wenn das Fuzzy-Set ein Trapez darstellt
	 */
	public boolean isTrapez() {
		if (isSpezialfall()) {
			return false;
		}

		return (t1 < t2) && (t2 < t3) && (t3 < t4);
	}

	/**
	 * Legt die erste Ecke des Trapez fest.
	 *
	 * @param t1
	 *            Erste Ecke des Trapez
	 * @throws IllegalArgumentException
	 *             , wenn die Trapezecken nicht aufsteigend sortiert sind
	 */
	public void setT1(final int t1) {
		if (!isSortiert(t1, t2, t3, t4) && !isSpezialfall(t1, t2, t3, t4)) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.TrapezeckenUnsortiert.toString());
		}

		this.t1 = t1;
	}

	/**
	 * Legt die zweite Ecke des Trapez fest.
	 *
	 * @param t2
	 *            Zweite Ecke des Trapez
	 * @throws IllegalArgumentException
	 *             , wenn die Trapezecken nicht aufsteigend sortiert sind
	 */
	public void setT2(final int t2) {
		if (!isSortiert(t1, t2, t3, t4) && !isSpezialfall(t1, t2, t3, t4)) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.TrapezeckenUnsortiert.toString());
		}

		this.t2 = t2;
	}

	/**
	 * Legt die dritte Ecke des Trapez fest.
	 *
	 * @param t3
	 *            Dritte Ecke des Trapez
	 * @throws IllegalArgumentException
	 *             , wenn die Trapezecken nicht aufsteigend sortiert sind
	 */
	public void setT3(final int t3) {
		if (!isSortiert(t1, t2, t3, t4) && !isSpezialfall(t1, t2, t3, t4)) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.TrapezeckenUnsortiert.toString());
		}

		this.t3 = t3;
	}

	/**
	 * Legt die vierte Ecke des Trapez fest.
	 *
	 * @param t4
	 *            Vierte Ecke des Trapez
	 * @throws IllegalArgumentException
	 *             , wenn die Trapezecken nicht aufsteigend sortiert sind
	 */
	public void setT4(final int t4) {
		if (!isSortiert(t1, t2, t3, t4) && !isSpezialfall(t1, t2, t3, t4)) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.TrapezeckenUnsortiert.toString());
		}

		this.t4 = t4;
	}

	/**
	 * Gibt ein Tupel der Form Termname[t1, t2, t3, t4] zur&uuml;ck.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return name + "[" + t1 + ", " + t2 + ", " + t3 + ", " + t4 + "]"; 
	}

	/**
	 * Pr&uuml;ft ob die Trapezecken ordnungsgem&auml;&szlig; aufsteigend
	 * sortiert sind.
	 *
	 * @param p1
	 *            Erste Ecke
	 * @param p2
	 *            Zweite Ecke
	 * @param p3
	 *            Dritte Ecke
	 * @param p4
	 *            Vierte Ecke
	 * @return Sind die Ecken korrekt angegeben?
	 */
	private static boolean isSortiert(final int p1, final int p2, final int p3,
			final int p4) {
		return ((p1 <= p2) && (p2 <= p3) && (p3 <= p4));
	}

	/**
	 * Pr&uuml;ft ob das Fuzzy-Set ein Spezialfall ist. Es gibt genau einen
	 * Spezialfall und zwar, wenn das Fuzzy-Sets beide Wertebereichsgrenzen
	 * schneidet. Dies ist der einzige erlaubte Fall, in dem die Trapezecken
	 * nicht aufsteigend sortiert sind.
	 *
	 * @param p1
	 *            Erste Ecke
	 * @param p2
	 *            Zweite Ecke
	 * @param p3
	 *            Dritte Ecke
	 * @param p4
	 *            Vierte Ecke
	 * @return Liegt der Spezialfall vor?
	 */
	private static boolean isSpezialfall(final int p1, final int p2,
			final int p3, final int p4) {
		return ((p1 < p2) && (p2 > p3) && (p3 < p4));
	}

}
